"""All (global/common) constants for Music Assistant."""

SECURE_STRING_SUBSTITUTE = "this_value_is_encrypted"

PLAYER_CONTROL_NATIVE = "native"
PLAYER_CONTROL_NONE = "none"
PLAYER_CONTROL_FAKE = "fake"
